/* This is the interface to the python interpreter
   Originally from the now defunct gnocci project.
   C. Steenberg 15 May 2000
*/

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gtkextra/gtkextra.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <string.h>
#include <Python.h>
#include <grammar.h>
#include <node.h>
#include <parsetok.h>
#include <errcode.h>
#include <compile.h>
#include <eval.h>
#include <marshal.h>
#ifdef WITH_NUMERIC_PYTHON
#include <arrayobject.h>
#endif
#include "../main.h"
#include "python_int.h"


PyObject *eval_dict, *main_o, *dict_o;
extern gint stdout_pipe[2];
gint status=0; /* 0=normal, 1=old line waiting */
extern grammar _PyParser_Grammar; /* From Python/graminit.c */
extern FILE *stdout_fp;
GString *com_buffer;

int python_init(gchar *progname)
{   Py_SetProgramName(progname);
    Py_Initialize();
    PyRun_SimpleString ("import sys");
    PyRun_SimpleString ("import os");
    eval_dict = Py_BuildValue("{s{}}", "__builtins__");
    main_o=PyImport_AddModule("__main__");
    dict_o=PyModule_GetDict(main_o);
/* Numerical python */
#ifdef WITH_NUMERIC_PYTHON
    PyRun_SimpleString ("from Numeric import *");
#endif
    return(1);
}

int python_singleton(SGworksheet *worksheet,gint row, gint col,
                     PyObject *object)
{ static PyObject *s_obj;
  GtkSheet *sheet;

  sheet = GTK_SHEET(worksheet->sheet);

  s_obj=PyObject_Str(object);

  if (s_obj)
   { Py_INCREF(s_obj);
     gtk_sheet_set_cell_text(GTK_SHEET(sheet),row,col,PyString_AsString (s_obj));
     Py_DECREF(s_obj);
   }
  return 1;
}

int python_sequence(SGworksheet *worksheet,gint row, gint col, PyObject *object,
                    gint depth)
{ gint len,i;
  PyObject *seq_obj;
  GtkSheet *sheet;

  if (depth>1) return 0; /* We handle only 2D lists for now */

  sheet = GTK_SHEET(worksheet->sheet);

  len=PySequence_Length(object);

  /* Check to see if we need to resize the sheet */
  seq_obj=PySequence_GetItem(object,0);
  if (PySequence_Check(seq_obj) && gtk_sheet_get_maxcol(sheet) <
      len+col+1)
       sg_worksheet_add_columns(worksheet,len-
            gtk_sheet_get_maxcol(sheet)+col-1);
  else if (!PySequence_Check(seq_obj) && gtk_sheet_get_maxrow(sheet) < row+len)
        sg_worksheet_add_rows(worksheet,len-gtk_sheet_get_maxrow(sheet)+row-1);

  for (i=0;i<len;i++)
    { seq_obj=PySequence_GetItem(object,i);
      Py_INCREF(seq_obj);
      if (PySequence_Check(seq_obj))
        python_sequence(worksheet,row,col+i,seq_obj,depth+1);
      else
        python_singleton(worksheet,row+i,col,seq_obj);

      Py_XDECREF(seq_obj);
    }
  return 1;
}

#ifdef WITH_NUMERIC_PYTHON
int python_array_print(SGworksheet *worksheet,gint row, gint col,
                       PyArrayObject *object,gint dim, gchar *data)
{ gchar *array,singleton[80],CHAR[]="%%%dc",UBYTE[]="%%%dhi",
        SHORT[]="%%%dhd",
        INT[]="%%%dd",LONG[]="%%%dld",FLOAT[]="%%1.%dg",CFLOAT[]="%%1.%dLg",
        format[20],*fpt,*apt;
  gint stride,i,j,vert,rowskip;

  
  array=data;
  if (dim>1)
   { rowskip=1;
     for (i=dim;i>0;i--)
       rowskip*=object->dimensions[i];
     stride=object->strides[dim];
     for (i=0;i<object->dimensions[i];i++)
       { python_array_print(worksheet,row+i*rowskip,col,object,dim-1,array);
         array+=stride;
       }
   }
  else
    { if (dim>0)
       { vert=object->dimensions[1];
         stride=object->strides[1];
       }
      else
       { vert=1;
         stride=0;
       }
  /* Now insert the cell values */

      array=data;
      switch (object->descr->type_num)
       { case PyArray_CHAR: fpt=CHAR;
          break;
         case PyArray_UBYTE: fpt=UBYTE;
          break;
         case PyArray_SBYTE:
         case PyArray_SHORT: fpt=SHORT;
          break;
         case PyArray_INT: fpt=INT;
          break;
         case PyArray_LONG: fpt=LONG;
          break;
         case PyArray_FLOAT:
         case PyArray_DOUBLE: fpt=FLOAT;
          break;
         case PyArray_CFLOAT:
         case PyArray_CDOUBLE: fpt=CFLOAT;
          break;
       }
       snprintf(format,20,fpt,worksheet->column[col].precision);
      for (j=0;j<vert;j++)
        { for (i=0;i<object->dimensions[0];i++)
            { apt=array+i*object->strides[0];
              switch (object->descr->type_num)
               { case PyArray_CHAR:
                 case PyArray_UBYTE:
                 case PyArray_SBYTE:
                  snprintf(singleton,80,format,*(char *)(array+i*object->strides[0]));
                  break;
                 case PyArray_SHORT:
                  snprintf(singleton,80,format,*(short *)(array+i*object->strides[0]));
                  break;
                 case PyArray_INT:
                  snprintf(singleton,80,format,*(int *)(array+i*object->strides[0]));
                  break;
                 case PyArray_LONG:
                  snprintf(singleton,80,format,*(long *)(array+i*object->strides[0]));
                  break;
                 case PyArray_FLOAT:
                  snprintf(singleton,80,format,*(float *)(array+i*object->strides[0]));
                  break;
                 case PyArray_DOUBLE:
                  snprintf(singleton,80,format,*(double *)(array+i*object->strides[0]));
                  break;
                 case PyArray_CFLOAT:
                 case PyArray_CDOUBLE:
                  snprintf(singleton,80,format,*(long double *)(array+i*object->strides[0]));
                  break;
                 case PyArray_OBJECT:
                  python_sequence(worksheet,row+j,col+i,
                   (PyObject *)(array+i*object->strides[0]),0);
                  break;
               }

            gtk_sheet_set_cell_text(GTK_SHEET(worksheet->sheet),row+i,col+j,strdup(singleton));
          }
          array+=stride;
        }
    }
  return 1;
}

int python_array(SGworksheet *worksheet,gint row, gint col, PyArrayObject *object)
{ gint len,i;
  GtkSheet *sheet;

  sheet = GTK_SHEET(worksheet->sheet);

  if (object->nd>2) return 0; /* No nore than 2 dimensions for now */

  /* Check to see if we need to resize the sheet */
  if (object->nd>1 && 
      gtk_sheet_get_maxcol(sheet)< object->dimensions[1]+col+1)
     sg_worksheet_add_columns(worksheet,object->dimensions[1]-
            gtk_sheet_get_maxcol(sheet)+col-1);
  len=1;
  for (i=0;i<object->nd;i++)
    if (i!=1) len=len*object->dimensions[i];

  if (gtk_sheet_get_maxrow(sheet) < row+len)
        sg_worksheet_add_rows(worksheet,len-gtk_sheet_get_maxrow(sheet)+row-1);

  python_array_print(worksheet,row,col,object,object->nd-1,object->data);

  return 1;
}

#endif

int python_sheet(SGworksheet *worksheet,gint row, gint col, gchar *command)
{ PyObject *object=NULL;
  gchar errmsg[]="Error!";
  gchar *msg;
  GtkSheet *sheet;
  gint retval;

  sheet = GTK_SHEET(worksheet->sheet);

  object=PyRun_String (command, Py_eval_input, dict_o, dict_o);
  if (!object)
   { if (PyErr_Occurred()) PyErr_Clear();
     PyRun_String (command, Py_single_input, dict_o, dict_o);
   }

  if (object)
   {  Py_INCREF(object);
#ifdef WITH_NUMERIC_PYTHON
      if (PyArray_Check(object))
       { gtk_sheet_freeze(sheet);
         retval=python_array(worksheet,row,col,(PyArrayObject *)object);
         gtk_sheet_thaw(sheet);
       }
      else
#endif
      if (PySequence_Check(object))
       { gtk_sheet_freeze(sheet);
         retval=python_sequence(worksheet,row,col,object,0);
         gtk_sheet_thaw(sheet);
       }

      else
        retval=python_singleton(worksheet,row,col,object);

      Py_XDECREF(object);
   }
  else
   retval=0;

  if (!retval)
   { if (PyErr_Occurred())
      { msg=errmsg;
        PyErr_Clear();
      }
     else msg=command;
     gtk_sheet_set_cell_text(GTK_SHEET(sheet),row,col,msg);
     return 0;
   }
  return 1;
}
